from .exceptions import Pyz3rException, AlttprFailedToRetrieve, AlttprFailedToGenerate
from . import misc, spoiler
from .rom import Rom
import aiohttp
import logging
import tempfile
import os
from pathlib import Path

from pyz3r import rom

class ALTTPR():
    def __init__(
        self,
        baseurl='https://alttpr.com',
        username=None,
        password=None,
    ):
        self.data = None
        self.hash_id = None
        self.settings = None
        self.baseurl = baseurl
        self.auth = aiohttp.BasicAuth(login=username, password=password) if username and password else None

    @classmethod
    async def generate(cls, settings, endpoint='/api/randomizer', **kwargs):
        seed = cls(**kwargs)

        seed.data = await seed.generate_game(settings, endpoint)
        seed.hash = seed.data['hash']

        return seed

    @classmethod
    async def retrieve(cls, hash_id, **kwargs):
        seed = cls(**kwargs)

        seed.hash = hash_id
        seed.data = await seed.retrieve_game(hash_id)

        return seed

    async def generate_game(self, settings, endpoint):
        self.settings = settings
        for _ in range(0, 5):
            try:
                async with aiohttp.request(method='post', url=self.uri(endpoint), json=settings, auth=self.auth, raise_for_status=True) as resp:
                    req = await resp.json()
                self.data = req
                return req
            except (aiohttp.client_exceptions.ServerDisconnectedError, aiohttp.ClientResponseError):
                logging.exception("Unable to generate game.")

        raise AlttprFailedToGenerate('failed to generate game')

    async def retrieve_game(self, hash_id):
        for _ in range(0, 5):
            try:
                async with aiohttp.request(method='get', url=self.uri('/hash/' + hash_id), auth=self.auth, raise_for_status=True) as resp:
                    req = await resp.json(content_type="text/html")
                self.data = req
                return req
            except aiohttp.ClientResponseError:
                logging.exception("Unable to retrieve game.")
            except aiohttp.client_exceptions.ServerDisconnectedError:
                logging.exception("Unable to retrieve game.")
        raise AlttprFailedToRetrieve(
            f'failed to retrieve game {hash_id}, the game is likely not found.')


    async def randomizer_settings(self):
        """Returns a dictonary of valid settings, based on the randomizer in use (item or entrance).

        Returns:
            dict -- dictonary of valid settings that can be used
        """
        async with aiohttp.request(method='get', url=self.baseurl + '/randomizer/settings', auth=self.auth, raise_for_status=True) as resp:
            settings = await resp.json()
        return settings

    async def customizer_settings(self):
        """Returns a dictonary of valid settings, based on the randomizer in use (item or entrance).

        Returns:
            dict -- dictonary of valid settings that can be used
        """
        async with aiohttp.request(method='get', url=self.baseurl + '/customizer/settings', auth=self.auth, raise_for_status=True) as resp:
            settings = await resp.json()
        return settings

    async def find_daily_hash(self):
        async with aiohttp.request(method='get', url=f'{self.baseurl}/api/daily', auth=self.auth, raise_for_status=True) as resp:
            daily = await resp.json()
        return daily['hash']

    @property
    def url(self):
        return f'{self.baseurl}/h/{self.hash}'

    @property
    def code(self):
        """An list of strings that represents the

        Raises:
            Pyz3rException -- Raised if no game has been generated or retrieved.

        Returns:
            list -- List of strings depicting the code on the in-game file select screen.
        """
        if not self.data:
            raise Pyz3rException('Please specify a seed or hash first to generate or retrieve a game.')

        code_map = {
            0: 'Bow', 1: 'Boomerang', 2: 'Hookshot', 3: 'Bombs',
            4: 'Mushroom', 5: 'Magic Powder', 6: 'Ice Rod', 7: 'Pendant',
            8: 'Bombos', 9: 'Ether', 10: 'Quake', 11: 'Lamp',
            12: 'Hammer', 13: 'Shovel', 14: 'Flute', 15: 'Bugnet', 16: 'Book',
            17: 'Empty Bottle', 18: 'Green Potion', 19: 'Somaria', 20: 'Cape',
            21: 'Mirror', 22: 'Boots', 23: 'Gloves', 24: 'Flippers',
            25: 'Moon Pearl', 26: 'Shield', 27: 'Tunic', 28: 'Heart',
            29: 'Map', 30: 'Compass', 31: 'Big Key'
        }

        codebytes = misc.seek_patch_data(self.data['patch'], 1573397, 5)
        if len(codebytes) != 5:
            return ["Bow", "Boomerang", "Hookshot", "Bombs", "Mushroom"]
        else:
            p = list(map(lambda x: code_map[x], codebytes))
            return [p[0], p[1], p[2], p[3], p[4]]

    async def get_patch_base(self):
        """Gets the base_rom from the website.  This is the first set of patches that must be applied to the ROM.
        Caches BPS files in the system's temp directory (/tmp/pyz3r/bps on *nix systems).

        Returns:
            bytes -- a bytes-like object representing a BPS patch
        """
        async with aiohttp.request(method='get', url=self.uri("/api/h/" + self.hash), auth=self.auth, raise_for_status=True) as resp:
            seed_settings = await resp.json()

        cachedbpstmp = os.path.join(tempfile.gettempdir(), "pyz3r", "bps")
        cachedbpsfile = os.path.join(cachedbpstmp, f"{seed_settings['md5']}.bps")
        try:
            Path(cachedbpstmp).mkdir(parents=True, exist_ok=True)
            with open(cachedbpsfile, "rb") as f:
                req_patch = f.read()
        except (FileNotFoundError, PermissionError):
            async with aiohttp.request(method='get', url=self.baseurl + seed_settings['bpsLocation'], auth=self.auth, raise_for_status=True) as resp:
                req_patch = await resp.read()

            try:
                with open(cachedbpsfile, "wb") as f:
                    f.write(req_patch)
            except PermissionError:
                logging.exception("Unable to cache BPS file.")

        return req_patch

    async def create_patched_game(
        self,
        input_filename,
        output_filename=None,
        heartspeed='half',
        heartcolor='red',
        quickswap=False,
        menu_speed='normal',
        spritename='Link',
        music=True) -> Rom:
        if not self.data:
            raise Pyz3rException('Please specify a seed or hash first to generate or retrieve a game.')

        self.rom = Rom(input_filename)

        self.rom.apply_bps_patch(patch=await self.get_patch_base())

        # expand the ROM to size requested in seed_data
        if self.data['size'] > 2:
            self.rom.expand(newlenmb=self.data['size'])

        # apply the seed-specific changes
        self.rom.apply_dict_patches(patches=self.data['patch'])

        # apply the heart speed change
        self.rom.heart_speed(heartspeed)

        # apply the heart color change
        self.rom.heart_color(heartcolor)

        # apply menu speed
        self.rom.menu_speed(menu_speed)

        # apply quickswap
        self.rom.quickswap(quickswap)

        if spritename != "Link":
            # apply the sprite
            self.rom.sprite(zspr=await self.get_sprite(spritename))

        # apply the sprite
        self.rom.music(music=music)

        # calculate the SNES checksum and apply it to the ROM
        self.rom.checksum()

        if output_filename is not None:
            self.rom.write_to_file(output_filename)

        return self.rom

    async def get_sprite(self, name):
        """Retrieve the ZSPR file for the named sprite.

        Arguments:
            name {str} -- The name of the sprite, as listed at https://alttpr.com/en/sprite_preview

        Raises:
            Pyz3rException -- Raised if no game has been generated or retrieved.

        Returns:
            list -- a list of bytes depicting a SPR or ZSPR file
        """
        async with aiohttp.request(method='get', url=self.baseurl + '/sprites', auth=self.auth, raise_for_status=True) as resp:
            sprites = await resp.json()
        try:
            spriteinfo = next(
                (sprite for sprite in sprites if sprite["name"] == name))
        except StopIteration:
            raise Pyz3rException(
                f"Sprite {name} does not exist on {self.base_url}.")
        try:
            async with aiohttp.request(method='get', url=spriteinfo["file"], raise_for_status=True) as resp:
                spritedata = await resp.read()
        except Exception as e:
            raise Pyz3rException(
                f'Sprite "{name}" could not be downloaded.') from e

        return bytearray(spritedata)

    def uri(self, url):
        return f'{self.baseurl}{url}'

    def get_formatted_spoiler(self, translate_dungeon_items=False):
        return spoiler.create_filtered_spoiler(self, translate_dungeon_items)



import bisect


def chunk(iterator, count):
    itr = iter(iterator)
    while True:
        try:
            yield tuple([next(itr) for i in range(count)])
        except StopIteration:
            return


def seek_patch_data(patches, offset, num_bytes):
    offsetlist = []
    for patch in patches:
        for key, value in patch.items():
            offsetlist.append(int(key))
    offsetlist_sorted = sorted(offsetlist)
    i = bisect.bisect_left(offsetlist_sorted, offset)
    if i:
        if offsetlist_sorted[i] == offset:
            seek = str(offset)
            for patch in patches:
                if seek in patch:
                    return patch[seek][:num_bytes]
        else:
            left_slice = offset - offsetlist_sorted[i - 1]
            for patch in patches:
                seek = str(offsetlist_sorted[i - 1])
                if seek in patch:
                    return patch[seek][left_slice:left_slice + num_bytes]
    raise ValueError


def convert_randomizer_settings(web_dict):
    """
    Convert web/JS randomizer presets to API compatible presets

    There is no API from ALTTPR website to provide API compatible generation
    configs.  There is, however, an API for web/JS compatible presets.  This
    method converts those presets so that they can be used with this library.
    """
    return {
            "glitches": web_dict.get("glitches_required"),
            "item_placement": web_dict.get("item_placement"),
            "dungeon_items": web_dict.get("dungeon_items"),
            "accessibility": web_dict.get("accessibility"),
            "goal": web_dict.get("goal"),
            "crystals": {
                "ganon": web_dict.get("ganon_open"),
                "tower": web_dict.get("tower_open"),
                },
            "mode": web_dict.get("world_state"),
            "entrances": web_dict.get("entrance_shuffle"),
            "hints": web_dict.get("hints"),
            "weapons": web_dict.get("weapons"),
            "item": {
                "pool": web_dict.get("item_pool"),
                "functionality": web_dict.get("item_functionality"),
                },
            "lang": "en",
            "enemizer": {
                "boss_shuffle": web_dict.get("boss_shuffle"),
                "enemy_shuffle": web_dict.get("enemy_shuffle"),
                "enemy_damage": web_dict.get("enemy_damage"),
                "enemy_health": web_dict.get("enemy_health"),
                }
           }

# shamelessly stolen from https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries
def mergedicts(dict1, dict2):
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(mergedicts(dict1[k], dict2[k])))
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                yield (k, dict2[k])
                # Alternatively, replace this with exception raiser to alert you of value conflicts
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


import slugid
import uuid
import asyncio
import aiohttp
from . import exceptions

async def sm(
    settings=None,
    slug_id=None,
    guid_id=None,
    baseurl='https://samus.link',
    randomizer='sm',
    username=None,
    password=None,
):
    seed = smClass(
        settings=settings,
        slug_id=slug_id,
        guid_id=guid_id,
        baseurl=baseurl,
        randomizer=randomizer,
        username=username,
        password=password
    )
    await seed._init()
    return seed


class smClass():
    def __init__(
        self,
        settings,
        slug_id,
        guid_id,
        baseurl,
        randomizer,
        username,
        password,
    ):
        self.settings = settings
        self.slug_id = slug_id
        self.guid_id = guid_id
        self.baseurl = baseurl
        self.randomizer = randomizer
        self.username = username
        self.password = password
        self.auth = aiohttp.BasicAuth(
            login=username, password=password) if username and password else None

    async def _init(self):
        if self.settings:
            self.endpoint = f'/api/randomizers/{self.randomizer}/generate'
            self.data = await self.generate_game()
            self.guid = uuid.UUID(hex=self.data['guid'])
            self.slug_id = slugid.encode(self.guid)
        elif self.slug_id:
            self.guid = slugid.decode(self.slug_id)
            self.data = await self.retrieve_game()
        elif self.guid_id:
            self.guid = uuid.UUID(hex=self.guid_id)
            self.slug_id = slugid.encode(self.guid)
            self.data = await self.retrieve_game()
        else:
            self.data = None
            self.slug_id = None
            self.guid = None

    async def generate_game(self):
        for i in range(0, 5):
            try:
                async with aiohttp.request(method='post', url=self.baseurl + self.endpoint, json=self.settings, auth=self.auth) as resp:
                    req = await resp.json()
            except aiohttp.client_exceptions.ServerDisconnectedError:
                continue
            # except aiohttp.ClientResponseError:
            #     continue
            return req
        raise exceptions.AlttprFailedToGenerate('failed to generate game')

    async def retrieve_game(self):
        for i in range(0, 5):
            try:
                async with aiohttp.request(method='get', url=f'{self.baseurl}/api/seed/{self.guid.hex}') as resp:
                    patch = await resp.json()
            except aiohttp.ClientResponseError:
                await asyncio.sleep(5)
                continue
            except aiohttp.client_exceptions.ServerDisconnectedError:
                continue
            return patch
        raise exceptions.AlttprFailedToRetrieve(
            f'failed to retrieve game {self.slug_id}, the game is likely not found')

    @classmethod
    async def create(
        cls,
        settings=None,
        slug_id=None,
        guid_id=None,
        baseurl='https://samus.link',
        randomizer='sm',
        username=None,
        password=None,
    ):
        seed = cls(
            settings=settings,
            slug_id=slug_id,
            guid_id=guid_id,
            baseurl=baseurl,
            randomizer=randomizer,
            username=username,
            password=password
        )
        await seed._init()
        return seed

    async def randomizer_settings(self):
        """Returns a dictonary of valid settings.

        Returns:
            dict -- dictonary of valid settings that can be used
        """
        async with aiohttp.request(method='get', url=f'{self.base_url}/api/randomizers/{self.randomizer}') as resp:
            settings = await resp.json()
        return settings

    @property
    def url(self):
        if self.data.get('mode', 'normal') == 'multiworld':
            return f'{self.baseurl}/multiworld/{self.slug_id}'
        return f'{self.baseurl}/seed/{self.slug_id}'

    @property
    def code(self):
        return self.data['hash']



class Pyz3rException(Exception):
    pass


class AlttprFailedToRetrieve(Pyz3rException):
    pass


class AlttprFailedToGenerate(Pyz3rException):
    pass


class UnableToRetrieve(Pyz3rException):
    pass

class UnableToGenerate(Pyz3rException):
    pass


import io
import itertools

import bps.apply

from .exceptions import Pyz3rException


class Rom(object):
    def __init__(self, input_filename):
        self.read(input_filename)

    def read(self, filename: str):
        with open(filename, "rb") as f:
            self.rom = bytearray(f.read())

    def write_to_file(self, path: str):
        with open(path, "wb") as f:
            f.write(self.rom)

    def apply_bps_patch(self, patch: bytes):
        """Applies a patch, which is a bps

        Arguments:
            patch {bytes} -- A series of bytes representing a bps patch
        """
        # create "files" for apply_to_files
        # although there is an apply_to_bytearrays it does not perform CRC
        # validation of input or output, apply_to_files does.
        patch_file = io.BytesIO(patch)
        source_file = io.BytesIO(bytes(self.rom))
        target_file = io.BytesIO()

        bps.apply.apply_to_files(patch_file, source_file, target_file)
        target_file.seek(0)
        self.rom = bytearray(target_file.read())

    def apply_dict_patches(self, patches: list):
        """Applies a patch, which is a list of dictionaries

        Arguments:
            patches {list} -- A list of dictionaries that depict of set of patches to be applied to the ROM.
        """
        for patch in patches:
            offset = int(list(patch.keys())[0])
            values = list(patch.values())[0]
            self.write_bytes(offset, values)

    def write_byte(self, offset, value):
        self.rom[offset] = value

    def write_bytes(self, offset, values):
        for idx, value in enumerate(values):
            self.write_byte(offset + idx, value)

    def heart_speed(self, speed='half'):
        """Set the low-health warning beep interval.

        Keyword Arguments:
            speed {str} -- Chose the speed at which the low health warning beeps.
                Options are 'off', 'double', 'normal', 'half', and 'quarter'. (default: {'half'})
        """
        sbyte = {
            'off': 0,
            'double': 16,
            'normal': 32,
            'half': 64,
            'quarter': 128,
        }
        self.write_byte(0x180033, sbyte[speed])

    def heart_color(self, color='red'):
        """Set the color of the hearts on the player's HUD.

        Keyword Arguments:
            color {str} -- The heart color.  Options are 'red', 'blue', 'green', and 'yellow' (default: {'red'})
        """

        cbyte = {
            'blue': [44, 13],
            'green': [60, 25],
            'yellow': [40, 9],
            'red': [36, 5],
        }

        byte = cbyte[color][0]
        file_byte = cbyte[color][1]

        self.write_byte(0x6fa1e, byte)
        self.write_byte(0x6fa20, byte)
        self.write_byte(0x6fa22, byte)
        self.write_byte(0x6fa24, byte)
        self.write_byte(0x6fa26, byte)
        self.write_byte(0x6fa28, byte)
        self.write_byte(0x6fa2a, byte)
        self.write_byte(0x6fa2c, byte)
        self.write_byte(0x6fa2e, byte)
        self.write_byte(0x6fa30, byte)
        self.write_byte(0x65561, file_byte)

    def music(self, music=True):
        """Enables, or disables, the in-game music.  Useful if you want to use an MSU-1 soundtrack instead.

        Keyword Arguments:
            music {bool} -- If true, music is enabled.  If false, the music id disabled. (default: {True})

        Returns:
            list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
        """

        self.write_byte(0x18021a, 0x00 if music else 0x01)

    def quickswap(self, quickswap=False):
        self.write_byte(0x18004b, 0x01 if quickswap else 0x00)

    def reduce_flashing(self, enable=False):
        self.write_byte(0x18017f, 0x01 if enable else 0x00)

    def menu_speed(self, speed='normal'):
        sbyte = {
            'instant': 0xE8,
            'fast': 0x10,
            'normal': 0x08,
            'slow': 0x04
        }
        self.write_byte(0x180048, sbyte[speed])
        self.write_byte(0x6dd9a, 0x20 if speed == 'instant' else 0x11)
        self.write_byte(0x6df2a, 0x20 if speed == 'instant' else 0x12)
        self.write_byte(0x6e0e9, 0x20 if speed == 'instant' else 0x12)

    def sprite(self, zspr):
        """Creates a patch for to replace Link's sprite with the contents of a ZSPR file.

        Arguments:
            spr {list} -- a bytes-like object of a ZSPR file
        """

        # stolen from VT's code
        gfx_offset = zspr[12] << 24 | zspr[11] << 16 | zspr[10] << 8 | zspr[9]
        palette_offset = zspr[18] << 24 | zspr[17] << 16 | zspr[16] << 8 | zspr[15]

        if self.rom[0x118000] == 0x02 and self.rom[0x118001] == 0x37 and self.rom[0x11801E] == 0x02 and self.rom[0x11801F] == 0x37:
            # skip past unicode title and author
            metadata_index = 0x1D
            junk = 2
            while metadata_index < gfx_offset and junk > 0:
                if zspr[metadata_index + 1] == 0 and zspr[metadata_index] == 0:
                    junk = junk - 1

                metadata_index = metadata_index + 2

            sprite_author_short = ""

            while metadata_index < gfx_offset and zspr[metadata_index] != 0x00:
                char = chr(zspr[metadata_index])
                sprite_author_short += char
                metadata_index = metadata_index + 1

            sprite_author_map = {
                " ": (0x9F, 0x9F), "0": (0x53, 0x79), "1": (0x54, 0x7A),
                "2": (0x55, 0x7B), "3": (0x56, 0x7C), "4": (0x57, 0x7D),
                "5": (0x58, 0x7E), "6": (0x59, 0x7F), "7": (0x5A, 0x80),
                "8": (0x5B, 0x81), "9": (0x5C, 0x82), "A": (0x5D, 0x83),
                "B": (0x5E, 0x84), "C": (0x5F, 0x85), "D": (0x60, 0x86),
                "E": (0x61, 0x87), "F": (0x62, 0x88), "G": (0x63, 0x89),
                "H": (0x64, 0x8A), "I": (0x65, 0x8B), "J": (0x66, 0x8C),
                "K": (0x67, 0x8D), "L": (0x68, 0x8E), "M": (0x69, 0x8F),
                "N": (0x6A, 0x90), "O": (0x6B, 0x91), "P": (0x6C, 0x92),
                "Q": (0x6D, 0x93), "R": (0x6E, 0x94), "S": (0x6F, 0x95),
                "T": (0x70, 0x96), "U": (0x71, 0x97), "V": (0x72, 0x98),
                "W": (0x73, 0x99), "X": (0x74, 0x9A), "Y": (0x75, 0x9B),
                "Z": (0x76, 0x9C), "'": (0x77, 0x9d), ".": (0xA0, 0xC0),
                "/": (0xA2, 0xC2), ":": (0xA3, 0xC3), "_": (0xA6, 0xC6)
            }

            sprite_author_short = sprite_author_short[:28].center(28, ' ').upper()
            sprite_author_bytes = [sprite_author_map[a] if a in sprite_author_map else (0x9F, 0x9F) for a in sprite_author_short]

            for idx, (upper, lower) in enumerate(sprite_author_bytes):
                self.write_byte(0x118002 + idx, upper)
                self.write_byte(0x118020 + idx, lower)

        self.write_bytes(0x80000, zspr[gfx_offset:gfx_offset + 28671])
        self.write_bytes(0xDD308, zspr[palette_offset:palette_offset + 120])
        self.write_bytes(0xDEDF5, zspr[palette_offset + 120:palette_offset + 120 + 4])

    def checksum(self):
        """Writes a patch that fixes a ROM's checksum.  This should be the last patch applied to a ROM before it is written.
        """

        sum_of_bytes = sum(self.rom[:32731]) + sum(self.rom[32736:])
        checksum = (sum_of_bytes + 510) & 65535
        inverse = checksum ^ 65535
        self.write_bytes(0x7fdc, [
            inverse & 255,
            inverse >> 8,
            checksum & 255,
            checksum >> 8,
        ])

    def expand(self, newlenmb):
        """Expands the byte list of a ROM to the specified number of megabytes, filling in the new space with zeroes.

        Arguments:
            rom {list} -- a list of bytes depicitng the rom

        Keyword Arguments:
            newlenmb {int} -- The size of the ROM should be, in megabytes.

        Raises:
            Pyz3rException -- Raised if the new length is shorter than the current size of the byte list.

        Returns:
            list -- a list of bytes depicitng the rom
        """

        newlen = int(newlenmb) * 1024 * 1024
        if len(self.rom) > newlen:
            raise Pyz3rException(f'ROM is already larger than {newlen}')
        diff = len(self.rom) - newlen
        if diff > 0:
            self.rom[newlen] = 0
        else:
            self.rom.extend(itertools.repeat(0, -diff))
            self.rom.append(0)



"""
ALTTPR API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the ALTTPR.com API.
"""

__title__ = 'pyz3r'
__author__ = 'Thomas Prescott'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018-2019 Thomas Prescott'
__version__ = '6.0.0'

from pyz3r.alttpr import ALTTPR
from pyz3r.sm import sm
from pyz3r.smvaria import SuperMetroidVaria
from pyz3r import mystery, customizer, rom